@@ -6,7 +6,7 @@
  * 
  * u_request.c: request related functions defintions
  * 
- * Copyright 2015-2017 Nicolas Mora <mail@babelouest.org>
+ * Copyright 2015-2020 Nicolas Mora <mail@babelouest.org>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -23,11 +23,17 @@
  * 
  */
 #include <stdlib.h>
+#include <stdarg.h>
 #include <string.h>
+#include <ctype.h>
 
 #include "u_private.h"
 #include "ulfius.h"
 
+#ifdef _MSC_VER
+#define strtok_r strtok_s
+#endif
+
 /**
  * Splits the url to an array of char *
  */
@@ -122,6 +128,42 @@ static int ulfius_url_format_match(const char ** splitted_url, const char ** spl
   return (splitted_url[i] == NULL && splitted_url_format[i] == NULL);
 }
 
+/**
+ * Converts a hex character to its integer value
+ */
+static char from_hex(char ch) {
+  return isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10;
+}
+
+/**
+ * Returns a url-decoded version of str
+ * IMPORTANT: be sure to free() the returned string after use
+ * Thanks Geek Hideout!
+ * http://www.geekhideout.com/urlcode.shtml
+ */
+static char * url_decode(const char * str) {
+  if (str != NULL) {
+    char * pstr = (char*)str, * buf = malloc(strlen(str) + 1), * pbuf = buf;
+    while (* pstr) {
+      if (* pstr == '%') {
+        if (pstr[1] && pstr[2]) {
+          * pbuf++ = from_hex(pstr[1]) << 4 | from_hex(pstr[2]);
+          pstr += 2;
+        }
+      } else if (* pstr == '+') { 
+        * pbuf++ = ' ';
+      } else {
+        * pbuf++ = * pstr;
+      }
+      pstr++;
+    }
+    * pbuf = '\0';
+    return buf;
+  } else {
+    return NULL;
+  }
+}
+
 /**
  * ulfius_endpoint_match
  * return the endpoint array matching the url called with the proper http method
@@ -147,7 +189,14 @@ struct _u_endpoint ** ulfius_endpoint_match(const char * method, const char * ur
           if (splitted_url_format != NULL && ulfius_url_format_match((const char **)splitted_url, (const char **)splitted_url_format)) {
             endpoint_returned = o_realloc(endpoint_returned, (count+2)*sizeof(struct _u_endpoint *));
             if (endpoint_returned != NULL) {
-              endpoint_returned[count] = (endpoint_list + i);
+              endpoint_returned[count] = o_malloc(sizeof(struct _u_endpoint));
+              if (endpoint_returned[count] != NULL) {
+                if (ulfius_copy_endpoint(endpoint_returned[count], (endpoint_list + i)) != U_OK) {
+                  y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error ulfius_copy_endpoint for endpoint_returned[%zu]", count);
+                }
+              } else {
+                y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for endpoint_returned[%zu]", count);
+              }
               endpoint_returned[count + 1] = NULL;
             } else {
               y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error reallocating memory for endpoint_returned");
@@ -181,12 +230,12 @@ struct _u_endpoint ** ulfius_endpoint_match(const char * method, const char * ur
  */
 int ulfius_parse_url(const char * url, const struct _u_endpoint * endpoint, struct _u_map * map, int check_utf8) {
   char * saveptr = NULL, * cur_word = NULL, * url_cpy = NULL, * url_cpy_addr = NULL;
-  char * saveptr_format = NULL, * saveptr_prefix = NULL, * cur_word_format = NULL, * url_format_cpy = NULL, * url_format_cpy_addr = NULL;
+  char * saveptr_format = NULL, * saveptr_prefix = NULL, * cur_word_format = NULL, * url_format_cpy = NULL, * url_format_cpy_addr = NULL, * concat_url_param = NULL;
 
   if (map != NULL && endpoint != NULL) {
     url_cpy = url_cpy_addr = o_strdup(url);
     url_format_cpy = url_format_cpy_addr = o_strdup(endpoint->url_prefix);
-    cur_word = strtok_r( url_cpy, ULFIUS_URL_SEPARATOR, &saveptr );
+    cur_word = url_decode(strtok_r( url_cpy, ULFIUS_URL_SEPARATOR, &saveptr ));
     if (endpoint->url_prefix != NULL && url_format_cpy == NULL) {
       y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for url_format_cpy");
     } else if (url_format_cpy != NULL) {
@@ -194,7 +243,8 @@ int ulfius_parse_url(const char * url, const struct _u_endpoint * endpoint, stru
     }
     while (cur_word_format != NULL && cur_word != NULL) {
       // Ignoring url_prefix words
-      cur_word = strtok_r( NULL, ULFIUS_URL_SEPARATOR, &saveptr );
+      o_free(cur_word);
+      cur_word = url_decode(strtok_r( NULL, ULFIUS_URL_SEPARATOR, &saveptr ));
       cur_word_format = strtok_r( NULL, ULFIUS_URL_SEPARATOR, &saveptr_prefix );
     }
     o_free(url_format_cpy_addr);
@@ -209,7 +259,7 @@ int ulfius_parse_url(const char * url, const struct _u_endpoint * endpoint, stru
     while (cur_word_format != NULL && cur_word != NULL) {
       if ((cur_word_format[0] == ':' || cur_word_format[0] == '@') && (!check_utf8 || utf8_check(cur_word) == NULL)) {
         if (u_map_has_key(map, cur_word_format+1)) {
-          char * concat_url_param = msprintf("%s,%s", u_map_get(map, cur_word_format+1), cur_word);
+          concat_url_param = msprintf("%s,%s", u_map_get(map, cur_word_format+1), cur_word);
           if (concat_url_param == NULL) {
             y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating resources for concat_url_param");
             o_free(url_cpy_addr);
@@ -229,9 +279,11 @@ int ulfius_parse_url(const char * url, const struct _u_endpoint * endpoint, stru
           }
         }
       }
-      cur_word = strtok_r( NULL, ULFIUS_URL_SEPARATOR, &saveptr );
+      o_free(cur_word);
+      cur_word = url_decode(strtok_r( NULL, ULFIUS_URL_SEPARATOR, &saveptr ));
       cur_word_format = strtok_r( NULL, ULFIUS_URL_SEPARATOR, &saveptr_format );
     }
+    o_free(cur_word);
     o_free(url_cpy_addr);
     o_free(url_format_cpy_addr);
     url_cpy_addr = NULL;
@@ -268,13 +320,23 @@ int ulfius_init_request(struct _u_request * request) {
     request->http_protocol = NULL;
     request->http_verb = NULL;
     request->http_url = NULL;
+    request->url_path = NULL;
     request->proxy = NULL;
+#if MHD_VERSION >= 0x00095208
+    request->network_type = U_USE_ALL;
+#endif
     request->timeout = 0L;
     request->check_server_certificate = 1;
+    request->check_server_certificate_flag = U_SSL_VERIFY_PEER|U_SSL_VERIFY_HOSTNAME;
+    request->check_proxy_certificate = 1;
+    request->check_proxy_certificate_flag = U_SSL_VERIFY_PEER|U_SSL_VERIFY_HOSTNAME;
+    request->follow_redirect = 0;
+    request->ca_path = NULL;
     request->client_address = NULL;
     request->binary_body = NULL;
     request->binary_body_length = 0;
-#ifndef U_DISABLE_WEBSOCKET
+    request->callback_position = 0;
+#ifndef U_DISABLE_GNUTLS
     request->client_cert = NULL;
     request->client_cert_file = NULL;
     request->client_key_file = NULL;
@@ -298,10 +360,12 @@ int ulfius_clean_request(struct _u_request * request) {
     o_free(request->http_protocol);
     o_free(request->http_verb);
     o_free(request->http_url);
+    o_free(request->url_path);
     o_free(request->proxy);
     o_free(request->auth_basic_user);
     o_free(request->auth_basic_password);
     o_free(request->client_address);
+    o_free(request->ca_path);
     u_map_clean_full(request->map_url);
     u_map_clean_full(request->map_header);
     u_map_clean_full(request->map_cookie);
@@ -317,7 +381,7 @@ int ulfius_clean_request(struct _u_request * request) {
     request->map_cookie = NULL;
     request->map_post_body = NULL;
     request->binary_body = NULL;
-#ifndef U_DISABLE_WEBSOCKET
+#ifndef U_DISABLE_GNUTLS
     gnutls_x509_crt_deinit(request->client_cert);
     o_free(request->client_cert_file);
     o_free(request->client_key_file);
@@ -354,18 +418,30 @@ int ulfius_copy_request(struct _u_request * dest, const struct _u_request * sour
     dest->http_protocol = o_strdup(source->http_protocol);
     dest->http_verb = o_strdup(source->http_verb);
     dest->http_url = o_strdup(source->http_url);
+    dest->url_path = o_strdup(source->url_path);
     dest->proxy = o_strdup(source->proxy);
+#if MHD_VERSION >= 0x00095208
+    dest->network_type = source->network_type;
+#endif
     dest->check_server_certificate = source->check_server_certificate;
+    dest->check_server_certificate_flag = source->check_server_certificate_flag;
+    dest->check_proxy_certificate = source->check_proxy_certificate;
+    dest->check_proxy_certificate_flag = source->check_proxy_certificate_flag;
+    dest->follow_redirect = source->follow_redirect;
+    dest->ca_path = o_strdup(source->ca_path);
     dest->timeout = source->timeout;
     dest->auth_basic_user = o_strdup(source->auth_basic_user);
     dest->auth_basic_password = o_strdup(source->auth_basic_password);
-    dest->client_address = o_malloc(sizeof(struct sockaddr));
+    dest->callback_position = source->callback_position;
     
-    if (dest->client_address != NULL) {
-      memcpy(dest->client_address, source->client_address, sizeof(struct sockaddr));
-    } else {
-      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating resources for dest->client_address");
-      ret = U_ERROR_MEMORY;
+    if (source->client_address != NULL) {
+      dest->client_address = o_malloc(sizeof(struct sockaddr));
+      if (dest->client_address != NULL) {
+        memcpy(dest->client_address, source->client_address, sizeof(struct sockaddr));
+      } else {
+        y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating resources for dest->client_address");
+        ret = U_ERROR_MEMORY;
+      }
     }
     
     if (ret == U_OK && u_map_clean(dest->map_url) == U_OK && u_map_init(dest->map_url) == U_OK) {
@@ -409,17 +485,22 @@ int ulfius_copy_request(struct _u_request * dest, const struct _u_request * sour
     }
     
     if (ret == U_OK) {
-      dest->binary_body_length = source->binary_body_length;
-      dest->binary_body = o_malloc(source->binary_body_length);
-      if (dest->binary_body != NULL) {
-        memcpy(dest->binary_body, source->binary_body, source->binary_body_length);
-      } else {
-        y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating resources for dest->binary_body");
-        ret = U_ERROR_MEMORY;
+      if (source->binary_body_length) {
+        dest->binary_body_length = source->binary_body_length;
+        dest->binary_body = o_malloc(source->binary_body_length);
+        if (dest->binary_body != NULL) {
+          memcpy(dest->binary_body, source->binary_body, source->binary_body_length);
+        } else {
+          y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating resources for dest->binary_body");
+          ret = U_ERROR_MEMORY;
+        }
       }
     }
 
-#ifndef U_DISABLE_WEBSOCKET
+#ifndef U_DISABLE_GNUTLS
+    dest->client_cert_file = o_strdup(source->client_cert_file);
+    dest->client_key_file = o_strdup(source->client_key_file);
+    dest->client_key_password = o_strdup(source->client_key_password);
     if (ret == U_OK && source->client_cert != NULL) {
       if (gnutls_x509_crt_init(&dest->client_cert) == 0) {
         char * str_cert = ulfius_export_client_certificate_pem(source);
@@ -434,13 +515,266 @@ int ulfius_copy_request(struct _u_request * dest, const struct _u_request * sour
       }
     }
 #endif
-    
     return ret;
   } else {
     return U_ERROR_PARAMS;
   }
 }
 
+/**
+ * ulfius_set_request_properties
+ * Set a list of properties to a request
+ * return U_OK on success
+ */
+int ulfius_set_request_properties(struct _u_request * request, ...) {
+  u_option option;
+  int ret = U_OK;
+  const char * str_key, * str_value;
+  size_t size_value;
+#ifndef U_DISABLE_JANSSON
+  json_t * j_value;
+#endif
+  va_list vl;
+
+  if (request != NULL) {
+    va_start(vl, request);
+    for (option = va_arg(vl, uint); option != U_OPT_NONE && ret == U_OK; option = va_arg(vl, uint)) {
+      switch (option) {
+        case U_OPT_HTTP_VERB:
+          str_value = va_arg(vl, const char *);
+          o_free(request->http_verb);
+          if (o_strlen(str_value)) {
+            request->http_verb = o_strdup(str_value);
+          } else {
+            request->http_verb = NULL;
+          }
+          break;
+        case U_OPT_HTTP_URL:
+          str_value = va_arg(vl, const char *);
+          o_free(request->http_url);
+          if (o_strlen(str_value)) {
+            request->http_url = o_strdup(str_value);
+          } else {
+            request->http_url = NULL;
+          }
+          break;
+        case U_OPT_HTTP_PROXY:
+          str_value = va_arg(vl, const char *);
+          o_free(request->proxy);
+          if (o_strlen(str_value)) {
+            request->proxy = o_strdup(str_value);
+          } else {
+            request->proxy = NULL;
+          }
+          break;
+#if MHD_VERSION >= 0x00095208
+        case U_OPT_NETWORK_TYPE:
+          request->network_type = (unsigned short)va_arg(vl, int);
+          break;
+#endif
+        case U_OPT_CHECK_SERVER_CERTIFICATE:
+          request->check_server_certificate = va_arg(vl, int);
+          break;
+        case U_OPT_CHECK_SERVER_CERTIFICATE_FLAG:
+          request->check_server_certificate_flag = va_arg(vl, int);
+          break;
+        case U_OPT_CHECK_PROXY_CERTIFICATE:
+          request->check_proxy_certificate = va_arg(vl, int);
+          break;
+        case U_OPT_CHECK_PROXY_CERTIFICATE_FLAG:
+          request->check_proxy_certificate_flag = va_arg(vl, int);
+          break;
+        case U_OPT_FOLLOW_REDIRECT:
+          request->follow_redirect = va_arg(vl, int);
+          break;
+        case U_OPT_CA_PATH:
+          str_value = va_arg(vl, const char *);
+          o_free(request->ca_path);
+          if (o_strlen(str_value)) {
+            request->ca_path = o_strdup(str_value);
+          } else {
+            request->ca_path = NULL;
+          }
+          break;
+        case U_OPT_TIMEOUT:
+          request->timeout = (unsigned long)va_arg(vl, int);
+          break;
+        case U_OPT_AUTH_BASIC_USER:
+          str_value = va_arg(vl, const char *);
+          o_free(request->auth_basic_user);
+          if (o_strlen(str_value)) {
+            request->auth_basic_user = o_strdup(str_value);
+          } else {
+            request->auth_basic_user = NULL;
+          }
+          break;
+        case U_OPT_AUTH_BASIC_PASSWORD:
+          str_value = va_arg(vl, const char *);
+          o_free(request->auth_basic_password);
+          if (o_strlen(str_value)) {
+            request->auth_basic_password = o_strdup(str_value);
+          } else {
+            request->auth_basic_password = NULL;
+          }
+          break;
+        case U_OPT_URL_PARAMETER:
+          str_key = va_arg(vl, const char *);
+          str_value = va_arg(vl, const char *);
+          ret = u_map_put(request->map_url, str_key, str_value);
+          break;
+        case U_OPT_HEADER_PARAMETER:
+          str_key = va_arg(vl, const char *);
+          str_value = va_arg(vl, const char *);
+          ret = u_map_put(request->map_header, str_key, str_value);
+          break;
+        case U_OPT_COOKIE_PARAMETER:
+          str_key = va_arg(vl, const char *);
+          str_value = va_arg(vl, const char *);
+          ret = u_map_put(request->map_cookie, str_key, str_value);
+          break;
+        case U_OPT_POST_BODY_PARAMETER:
+          str_key = va_arg(vl, const char *);
+          str_value = va_arg(vl, const char *);
+          ret = u_map_put(request->map_post_body, str_key, str_value);
+          break;
+        case U_OPT_URL_PARAMETER_REMOVE:
+          str_value = va_arg(vl, const char *);
+          ret = u_map_remove_from_key(request->map_url, str_value);
+          break;
+        case U_OPT_HEADER_PARAMETER_REMOVE:
+          str_value = va_arg(vl, const char *);
+          ret = u_map_remove_from_key(request->map_header, str_value);
+          break;
+        case U_OPT_COOKIE_PARAMETER_REMOVE:
+          str_value = va_arg(vl, const char *);
+          ret = u_map_remove_from_key(request->map_cookie, str_value);
+          break;
+        case U_OPT_POST_BODY_PARAMETER_REMOVE:
+          str_value = va_arg(vl, const char *);
+          ret = u_map_remove_from_key(request->map_post_body, str_value);
+          break;
+        case U_OPT_BINARY_BODY:
+          str_value = va_arg(vl, const char *);
+          size_value = va_arg(vl, size_t);
+          ret = ulfius_set_binary_body_request(request, str_value, size_value);
+          break;
+        case U_OPT_STRING_BODY:
+          str_value = va_arg(vl, const char *);
+          ret = ulfius_set_string_body_request(request, str_value);
+          break;
+#ifndef U_DISABLE_JANSSON
+        case U_OPT_JSON_BODY:
+          j_value = va_arg(vl, json_t *);
+          ret = ulfius_set_json_body_request(request, j_value);
+          break;
+#endif
+#ifndef U_DISABLE_GNUTLS
+        case U_OPT_CLIENT_CERT_FILE:
+          str_value = va_arg(vl, const char *);
+          o_free(request->client_cert_file);
+          if (o_strlen(str_value)) {
+            request->client_cert_file = o_strdup(str_value);
+          } else {
+            request->client_cert_file = NULL;
+          }
+          break;
+        case U_OPT_CLIENT_KEY_FILE:
+          str_value = va_arg(vl, const char *);
+          o_free(request->client_key_file);
+          if (o_strlen(str_value)) {
+            request->client_key_file = o_strdup(str_value);
+          } else {
+            request->client_key_file = NULL;
+          }
+          break;
+        case U_OPT_CLIENT_KEY_PASSWORD:
+          str_value = va_arg(vl, const char *);
+          o_free(request->client_key_password);
+          if (o_strlen(str_value)) {
+            request->client_key_password = o_strdup(str_value);
+          } else {
+            request->client_key_password = NULL;
+          }
+          break;
+#endif
+        default:
+          ret = U_ERROR_PARAMS;
+          break;
+      }
+    }
+    va_end(vl);
+  } else {
+    y_log_message(Y_LOG_LEVEL_DEBUG, "Ulfius - Error input parameter");
+    ret = U_ERROR_PARAMS;
+  }
+  return ret;
+}
+
+/**
+ * ulfius_set_string_body_request
+ * Set a string string_body to a request
+ * string_body must end with a '\0' character
+ * return U_OK on success
+ */
+int ulfius_set_string_body_request(struct _u_request * request, const char * string_body) {
+  if (request != NULL && string_body != NULL) {
+    // Free all the bodies available
+    o_free(request->binary_body);
+    request->binary_body = o_strdup(string_body);
+    if (request->binary_body == NULL) {
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for request->binary_body");
+      return U_ERROR_MEMORY;
+    } else {
+      request->binary_body_length = o_strlen(string_body);
+      return U_OK;
+    }
+  } else {
+    return U_ERROR_PARAMS;
+  }
+}
+
+/**
+ * ulfius_set_binary_body_request
+ * Add a binary binary_body to a request
+ * return U_OK on success
+ */
+int ulfius_set_binary_body_request(struct _u_request * request, const char * binary_body, const size_t length) {
+  if (request != NULL && binary_body != NULL && length > 0) {
+    // Free all the bodies available
+    o_free(request->binary_body);
+    request->binary_body = NULL;
+    request->binary_body_length = 0;
+
+    request->binary_body = o_malloc(length);
+    if (request->binary_body == NULL) {
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for request->binary_body");
+      return U_ERROR_MEMORY;
+    }
+    memcpy(request->binary_body, binary_body, length);
+    request->binary_body_length = length;
+    return U_OK;
+  } else {
+    return U_ERROR_PARAMS;
+  }
+}
+
+/**
+ * ulfius_set_empty_body_request
+ * Set an empty request body
+ * return U_OK on success
+ */
+int ulfius_set_empty_body_request(struct _u_request * request) {
+  if (request != NULL) {
+    // Free all the bodies available
+    o_free(request->binary_body);
+    request->binary_body = NULL;
+    request->binary_body_length = 0;
+    return U_OK;
+  } else {
+    return U_ERROR_PARAMS;
+  }
+}
+
 /**
  * create a new request based on the source elements
  * returned value must be free'd after use
@@ -454,80 +788,18 @@ struct _u_request * ulfius_duplicate_request(const struct _u_request * request)
       return NULL;
     }
     if (ulfius_init_request(new_request) == U_OK) {
-      new_request->http_protocol = o_strdup(request->http_protocol);
-      new_request->http_verb = o_strdup(request->http_verb);
-      new_request->http_url = o_strdup(request->http_url);
-      new_request->proxy = o_strdup(request->proxy);
-      if ((new_request->http_verb == NULL && request->http_verb != NULL) ||
-          (new_request->http_url == NULL && request->http_url != NULL) ||
-          (new_request->proxy == NULL && request->proxy != NULL) ||
-          (new_request->http_protocol == NULL && request->http_protocol != NULL)) {
-        y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for ulfius_duplicate_request");
+      if (ulfius_copy_request(new_request, request) != U_OK) {
+        y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error ulfius_copy_request");
         ulfius_clean_request_full(new_request);
-        return NULL;
+        new_request = NULL;
       }
-      if (request->client_address != NULL) {
-        new_request->client_address = o_malloc(sizeof(struct sockaddr));
-        if (new_request->client_address == NULL) {
-          y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for new_request->client_address");
-          ulfius_clean_request_full(new_request);
-          return NULL;
-        }
-        memcpy(new_request->client_address, request->client_address, sizeof(struct sockaddr));
-      }
-      new_request->check_server_certificate = request->check_server_certificate;
-      new_request->timeout = request->timeout;
-      new_request->auth_basic_user = o_strdup(request->auth_basic_user);
-      new_request->auth_basic_password = o_strdup(request->auth_basic_password);
-      u_map_clean_full(new_request->map_url);
-      u_map_clean_full(new_request->map_header);
-      u_map_clean_full(new_request->map_cookie);
-      u_map_clean_full(new_request->map_post_body);
-      new_request->map_url = u_map_copy(request->map_url);
-      new_request->map_header = u_map_copy(request->map_header);
-      new_request->map_cookie = u_map_copy(request->map_cookie);
-      new_request->map_post_body = u_map_copy(request->map_post_body);
-      if ((new_request->map_url == NULL && request->map_url != NULL) || 
-          (new_request->map_header == NULL && request->map_header != NULL) || 
-          (new_request->map_cookie == NULL && request->map_cookie != NULL) || 
-          (new_request->map_post_body == NULL && request->map_post_body != NULL)) {
-        ulfius_clean_request_full(new_request);
-        return NULL;
-      }
-      if (request->binary_body != NULL && request->binary_body_length > 0) {
-        new_request->binary_body = o_malloc(request->binary_body_length);
-        if (new_request->binary_body == NULL) {
-          y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error allocating memory for new_request->binary_body");
-          ulfius_clean_request_full(new_request);
-          return NULL;
-        }
-        memcpy(new_request->binary_body, request->binary_body, request->binary_body_length);
-      } else {
-        new_request->binary_body_length = 0;
-        new_request->binary_body = NULL;
-      }
-      new_request->binary_body_length = request->binary_body_length;
-#ifndef U_DISABLE_WEBSOCKET
-      if (request->client_cert != NULL) {
-        if (gnutls_x509_crt_init(&new_request->client_cert) == 0) {
-          char * str_cert = ulfius_export_client_certificate_pem(request);
-          if (ulfius_import_client_certificate_pem(new_request, str_cert) != U_OK) {
-            y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error ulfius_import_client_certificate_pem");
-            ulfius_clean_request_full(new_request);
-            return NULL;
-          }
-          o_free(str_cert);
-        } else {
-          y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_init");
-          ulfius_clean_request_full(new_request);
-          return NULL;
-        }
-      }
-#endif
     } else {
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error ulfius_init_request");
       o_free(new_request);
       new_request = NULL;
     }
+  } else {
+    y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error source request is NULL");
   }
   return new_request;
 }
@@ -586,7 +858,7 @@ json_t * ulfius_get_json_body_request(const struct _u_request * request, json_er
 }
 #endif
 
-#ifndef U_DISABLE_WEBSOCKET
+#ifndef U_DISABLE_GNUTLS
 /*
  * ulfius_export_client_certificate_pem
  * Exports the client certificate using PEM format
@@ -596,13 +868,14 @@ json_t * ulfius_get_json_body_request(const struct _u_request * request, json_er
 char * ulfius_export_client_certificate_pem(const struct _u_request * request) {
   char * str_cert = NULL;
   gnutls_datum_t g_cert;
+  int ret = 0;
 
   if (request != NULL && request->client_cert != NULL) {
-    if (gnutls_x509_crt_export2(request->client_cert, GNUTLS_X509_FMT_PEM, &g_cert) == 0) {
+    if ((ret = gnutls_x509_crt_export2(request->client_cert, GNUTLS_X509_FMT_PEM, &g_cert)) >= 0) {
       str_cert = o_strndup((const char *)g_cert.data, g_cert.size);
       gnutls_free(g_cert.data);
     } else {
-      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_export2");
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_export2: %s", gnutls_strerror(ret));
     }
   }
 
@@ -617,19 +890,19 @@ char * ulfius_export_client_certificate_pem(const struct _u_request * request) {
  * return U_OK on success;
  */
 int ulfius_import_client_certificate_pem(struct _u_request * request, const char * str_cert) {
-  int ret;
+  int ret, res;
   gnutls_datum_t g_cert;
 
   if (request != NULL && str_cert != NULL) {
     g_cert.data = (unsigned char *)str_cert;
     g_cert.size = o_strlen(str_cert);
-    if (gnutls_x509_crt_init(&request->client_cert)) {
-      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_init");
+    if ((res = gnutls_x509_crt_init(&request->client_cert))) {
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_init: %s", gnutls_strerror(res));
       ret = U_ERROR;
-    } else if (gnutls_x509_crt_import(request->client_cert, &g_cert, GNUTLS_X509_FMT_PEM) == 0) {
+    } else if ((res = gnutls_x509_crt_import(request->client_cert, &g_cert, GNUTLS_X509_FMT_PEM)) >= 0) {
       ret = U_OK;
     } else {
-      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_import");
+      y_log_message(Y_LOG_LEVEL_ERROR, "Ulfius - Error gnutls_x509_crt_import: %s", gnutls_strerror(res));
       ret = U_ERROR;
     }
   } else {
