@@ -5,8 +5,17 @@
 #include <string.h>
 #include <errno.h>
 #include <time.h>
+#include <ctype.h>
+#include <sys/socket.h>
+#include <sys/select.h>
+#include <sys/time.h>
+#include <sys/types.h>
+
 #ifndef _WIN32
+  #include <sys/socket.h>
   #include <netinet/in.h>
+  #include <arpa/inet.h>
+  #include <netdb.h>
 #else
   #include <unistd.h>
 #endif
@@ -15,89 +24,238 @@
 #include <check.h>
 #include <ulfius.h>
 
-#define CERT_KEY "-----BEGIN PRIVATE KEY-----\
-MIIEvwIBADANBgkqhkiG9w0BAQEFAASCBKkwggSlAgEAAoIBAQDr90HrswgEmln/\
-rXeNqYq0boIvas5wu27hmeHDdGGKtkCWIWGAo9GUy45xqsI4mDl3bOWS+pmb/3yi\
-+nhe+BmYHvEqUFo1JfUcVMxaNEbdd9REytMjKdOS+kkLf++BBRoZI/g8DggIu+Ri\
-dOSypk+pUECyQxROsyCrB/FgXuKbyC4QNl7fqZxMSpzw7jsWCZiwFv4pu8kMqzDG\
-2wTl/r/4STyK4Pj2TVa/JVzbZbH7VfcjT8MdMsXvKhlmPywjbqo70Hnmt3cnakYF\
-X+07ncx/5mjYYd3eSFgiNXr7WNw2rhFKtfTUcjrqSw9FDxmHFWUU76mwJyUo02N9\
-ViakSoQpAgMBAAECggEBAJp/VBwdJpzM6yxqyaJpZbXpvTeKuQw6zMjN1nIBG3SV\
-DAjAZnSxziGcffGSmoQvt0CoflAT4MuxJkwXrwSPcUKWz9Sis82kwq4AH6TYIaYU\
-NVmtazzUwAC1+2maJJjXXFUlpfy8Oypsy4ZjfvIxzmrPbuzI2t0Ej9kr5DDzL3BL\
-CWQ/U7w7y4KC0Pnq1ueIzM+UJIfvI0ldUcXHWsAnjyQzwgFBC35qDOfDTw0YUJv+\
-ElfFFcGYCA+9wlQyhM/zhAWqKgZ2mwAS6WykgbSc7j4NDjlmZwf4ZuTxbDUV1kBX\
-pPH21snqO42CFpw9hRUAA0W0XydCIfUhH8/6tH9enQECgYEA+rM9f6cUk3c7aLWs\
-hnauVqJuyGhgCkMyF9sSxgfcs87OVLNuGgaTIfwcT/7oxAY8G7sY44cbk1ZRhh7y\
-6kf01xqiJeXxBQei1qiJxMb2gukvpeY81s2Mg9og5d9qbEhLzp8TdiRJHxLIiGwF\
-xOM69CpugKN4T0Zum7EBGeSvmBECgYEA8PRG5SRTE4JwzGtLuTbMbjYTqyEjXAkB\
-zo33a92znA0EXEeLCl845EUgzUkSkeN/T/uyWRjj0hrPU99UaaXHt3bc+lrDHrc7\
-WFAR3QoAfFFJPIqqwiHcBDdTeAozQ8IOqFIxspl72RukuRdeQR3EdfcF9TUZyUbU\
-k8SuRioggpkCgYA2scgnA3KvwYGKlKgxJc9fQ0zcGDlrw8E4BymPXsO9zs6hGAxb\
-TTfoYDJlGX361kli22zQpvdTK6/ZjQL+LfiyvTLHBeWRbVsPbfGwpp+9a9ZjYVnA\
-m1OeqIYo4Jc9TICNcZMzYTM6vkRVzwtrKw//mQpGsmNbGEilWvaciZHtoQKBgQDo\
-FDBQtir6SJIConm9/ETtBlLtai6Xj+lYnK6qC1DaxkLj6tjF9a9jVh3g/DfRopBW\
-ZnSCkpGkJcR54Up5s35ofCkdTdxPsmaLihuaje6nztc+Y8VS1LAIs41GunRkF/5s\
-KzbI8kIyfAitag+Toms+v93SLwIWNo27gh3lYOANSQKBgQDIidSO3fzB+jzJh7R0\
-Yy9ADWbBsLxc8u+sBdxmZBGl+l4YZWNPlQsnsafwcpJWT3le6N7Ri3iuOZw9KiGe\
-QDkc7olxUZZ3pshg+cOORK6jVE8v6FeUlLnxpeAWa4C4JDawGPTOBct6bVBl5sxi\
-7GaqDcEK1TSxc4cUaiiPDNNXQA==\
------END PRIVATE KEY-----"
-#define CERT_PEM "-----BEGIN CERTIFICATE-----\
-MIIDhTCCAm2gAwIBAgIJANrO2RnCbURLMA0GCSqGSIb3DQEBCwUAMFkxCzAJBgNV\
-BAYTAkFVMRMwEQYDVQQIDApTb21lLVN0YXRlMSEwHwYDVQQKDBhJbnRlcm5ldCBX\
-aWRnaXRzIFB0eSBMdGQxEjAQBgNVBAMMCWxvY2FsaG9zdDAeFw0xNzA0MjgxNTA0\
-NDVaFw0xODA0MjgxNTA0NDVaMFkxCzAJBgNVBAYTAkFVMRMwEQYDVQQIDApTb21l\
-LVN0YXRlMSEwHwYDVQQKDBhJbnRlcm5ldCBXaWRnaXRzIFB0eSBMdGQxEjAQBgNV\
-BAMMCWxvY2FsaG9zdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAOv3\
-QeuzCASaWf+td42pirRugi9qznC7buGZ4cN0YYq2QJYhYYCj0ZTLjnGqwjiYOXds\
-5ZL6mZv/fKL6eF74GZge8SpQWjUl9RxUzFo0Rt131ETK0yMp05L6SQt/74EFGhkj\
-+DwOCAi75GJ05LKmT6lQQLJDFE6zIKsH8WBe4pvILhA2Xt+pnExKnPDuOxYJmLAW\
-/im7yQyrMMbbBOX+v/hJPIrg+PZNVr8lXNtlsftV9yNPwx0yxe8qGWY/LCNuqjvQ\
-eea3dydqRgVf7TudzH/maNhh3d5IWCI1evtY3DauEUq19NRyOupLD0UPGYcVZRTv\
-qbAnJSjTY31WJqRKhCkCAwEAAaNQME4wHQYDVR0OBBYEFPFfmGA3jO9koBZNGNZC\
-T/dZHZyHMB8GA1UdIwQYMBaAFPFfmGA3jO9koBZNGNZCT/dZHZyHMAwGA1UdEwQF\
-MAMBAf8wDQYJKoZIhvcNAQELBQADggEBAIc8Yuom4vz82izNEV+9bcCvuabcVwLH\
-Qgpv5Nzy/W+1hDoqfMfKNwOSdUB7jZoDaNDG1WhjKGGCLTAx4Hx+q1LwUXvu4Bs1\
-woocge65bl85h10l2TxxnlT5BIJezm5r3NiZSwOK2zxxIEyL4vh+b/xqQblBEkR3\
-e4/A4Ugn9Egh8GdpF4klGp4MjjpRyAVI7BDaleAhvDSfPmm7ylHJ2y7CLI9ApOQY\
-glwRuTmowAZQtaSiE1Ox7QtWj858HDzzTZyFWRG/MNqQptn7AMTPJv3DivNfDNPj\
-fYxFAheH3CjryHqqR9DD+d9396W8mqEaUp+plMwSjpcTDSR4rEQkUJg=\
------END CERTIFICATE-----"
-#define CERT_CA "-----BEGIN CERTIFICATE-----\
-MIIF9zCCA9+gAwIBAgIUQhoxaAC0kHeq1DNRL8fA0/WZCsYwDQYJKoZIhvcNAQEL\
-BQAwgYoxCzAJBgNVBAYTAkNBMQ8wDQYDVQQIDAZRdWViZWMxDzANBgNVBAcMBlF1\
-ZWJlYzEPMA0GA1UECgwGVWxmaXVzMRAwDgYDVQQLDAd0ZXN0LWNhMRIwEAYDVQQD\
-DAlsb2NhbGhvc3QxIjAgBgkqhkiG9w0BCQEWE3dlYm1hc3RlckBsb2NhbGhvc3Qw\
-HhcNMTgxMjAyMTkzNTA1WhcNMTkxMjAyMTkzNTA1WjCBijELMAkGA1UEBhMCQ0Ex\
-DzANBgNVBAgMBlF1ZWJlYzEPMA0GA1UEBwwGUXVlYmVjMQ8wDQYDVQQKDAZVbGZp\
-dXMxEDAOBgNVBAsMB3Rlc3QtY2ExEjAQBgNVBAMMCWxvY2FsaG9zdDEiMCAGCSqG\
-SIb3DQEJARYTd2VibWFzdGVyQGxvY2FsaG9zdDCCAiIwDQYJKoZIhvcNAQEBBQAD\
-ggIPADCCAgoCggIBALG4KdvHcAsKqaNe4RPVvGUA7P+N9+8vs8dTBQ16VEUGHCp5\
-bWdqpvWdiHfxKmCZVtVUkYGwOsFmsbOsAfQgF1gcqTFyIRCwdt2PbvWso2n3S+aJ\
-gCdqzYMpCoW09/1izAGqXVWrWhtBEFJiFSx0wbkzEF87ATLv1eox0g1C3y9LuIv0\
-T1nK36jPzC/S7hazCd4IXlzPMy858VJn8qv09TeexYLEEcvn70vTQzoOKQKAWTDO\
-BgwSDcfmez2uXktIuJR7EW3weaj6xJQWpnB8eb+A3H82TtP/EhB20ipc/X/lNBaj\
-58yKj5XJwVAoteIGYDV7PBzwFhyqhbErDNsMDcR2CrNLIzpdG/BpIwVbAyjoYCUm\
-OGxw8k+WXfyvTyS9TTi//sZbNRaVGAir63LFwCh0YbJ36DfKpjPrUoqDkO93iX3C\
-x12Nq5Mry8v6PgQsPt+sqZsDYW2BU0ib/pb4cksgBZakJcMFLcYi5tlCqFis1ehn\
-QjAZWn+kGCc4Tmy/l9mBBdRiYpP77QH6QVh4bPHzxaYiOoLfBajE6dWCXdLYOu+m\
-oZQmlcRWuP5AhKi0QAaI1/RAKGpO0dEUH7Dnl5TqR8twC88ZetkZPEJYHb3Vsy7U\
-aA0EvoHi1xTdh0tPN7AL0RRD8ekyh273CUMHg3vI0qjv29MWSBzCDa9X2SL5AgMB\
-AAGjUzBRMB0GA1UdDgQWBBQkE2nQu0XaDkJdLSQsgjTIXULLGDAfBgNVHSMEGDAW\
-gBQkE2nQu0XaDkJdLSQsgjTIXULLGDAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3\
-DQEBCwUAA4ICAQASrWMYrsMtIfUsN3srEAZqutcVAN7PWbzVNAjWXEqbw/HV7m8J\
-w09MTfmwmQEhw4d7fwzfax5vqKeqfRb7i8f+HzR8b9egCn8n4hWUjiAhS1Naj8OP\
-RCrvPyVI1/h3iMZZXlITljtgk7rZn5Yv/WjKz/l32KIHwMzmk5aTGkmr0oyWxj1Y\
-eJymZ0RYih+X4hUvAlW3a3AlKbvq7KW9bYU4LHJWixgP16tXOR8k0aPKpyVuNCzN\
-7/23vfPeawQuW7C23+IuDuPyrpivw4dbcKNcy1KfGXmc8kQDIv6uPqdWa5ngvVua\
-SHb23nA4HG3k5ed8iqarDmu4nU9TTu8qTaMKSuEAjRsPnehO/Fi/m3mMSnGEXxa1\
-WWl1wZxvFoD4IsexNZaZcii9f0Ctn3KbL7HwlJliogc2jDVHheS5aVsVnl2UsNJf\
-x4nYzPZ6bdLawcviwo9eQWXlEjoKYyBTL1LFNt9S/iEW3WUOaaLJVH7+uHOn4/6N\
-wiMh3OqcyePyB+MKr3tvsmYZyYVwCW/pRmATM+F4ol23OWhXyYN+AFdXiMBjwdUX\
-jAwgDammq/ymxfgKE7IX8z9+f+8UVcBaiXVUTWzgSHnjjPW8+9WExUws8BFD7+81\
-G66c7c4qxP2fq5vQiYJUVEpNd4Z4+EbvMDrg4CsTVaoI1OWlHWcBfNHupw==\
------END CERTIFICATE-----"
+#define SMTP_FROM "sender@localhost"
+#define SMTP_TO "recipient@localhost"
+#define SMTP_HOST "localhost"
+#define SMTP_PORT 2525
+#define SMTP_SUBJECT "E-mail subject for the test"
+#define SMTP_BODY "E-mail body for the test as well"
+
+#define BACKLOG_MAX  (10)
+#define BUF_SIZE  4096
+#define STREQU(a,b)  (strcmp(a, b) == 0)
+
+#define BODY_NOT_REDIRECTED "This is the blue pill"
+#define BODY_REDIRECTED "Welcome to the Matrix, Neo!"
+
+struct smtp_manager {
+  char * mail_data;
+  unsigned int port;
+  int sockfd;
+};
+
+/**
+ * 
+ * Function that emulates a very simple SMTP server
+ * Taken from Kenneth Finnegan's ccsmtp program
+ * https://gist.github.com/PhirePhly/2914635
+ * This function is under the GPL2 license
+ * 
+ */
+static void handle_smtp (struct smtp_manager * manager) {
+  int rc, i;
+  char buffer[BUF_SIZE], bufferout[BUF_SIZE];
+  int buffer_offset = 0;
+  buffer[BUF_SIZE-1] = '\0';
+
+  // Flag for being inside of DATA verb
+  int inmessage = 0;
+
+  sprintf(bufferout, "220 ulfius.tld SMTP CCSMTP\r\n");
+  send(manager->sockfd, bufferout, strlen(bufferout), 0);
+  //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+
+  while (1) {
+    fd_set sockset;
+    struct timeval tv;
+
+    FD_ZERO(&sockset);
+    FD_SET(manager->sockfd, &sockset);
+    tv.tv_sec = 120; // Some SMTP servers pause for ~15s per message
+    tv.tv_usec = 0;
+
+    // Wait tv timeout for the server to send anything.
+    select(manager->sockfd+1, &sockset, NULL, NULL, &tv);
+
+    if (!FD_ISSET(manager->sockfd, &sockset)) {
+      y_log_message(Y_LOG_LEVEL_DEBUG, "%d: Socket timed out", manager->sockfd);
+      break;
+    }
+
+    int buffer_left = BUF_SIZE - buffer_offset - 1;
+    if (buffer_left == 0) {
+      y_log_message(Y_LOG_LEVEL_DEBUG, "%d: Command line too long", manager->sockfd);
+      sprintf(bufferout, "500 Too long\r\n");
+      send(manager->sockfd, bufferout, strlen(bufferout), 0);
+      //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      buffer_offset = 0;
+      continue;
+    }
+
+    rc = recv(manager->sockfd, buffer + buffer_offset, buffer_left, 0);
+    if (rc == 0) {
+      y_log_message(Y_LOG_LEVEL_DEBUG, "%d: Remote host closed socket", manager->sockfd);
+      break;
+    }
+    if (rc == -1) {
+      y_log_message(Y_LOG_LEVEL_DEBUG, "%d: Error on socket", manager->sockfd);
+      break;
+    }
+
+    buffer_offset += rc;
+
+    char *eol;
+
+    // Only process one line of the received buffer at a time
+    // If multiple lines were received in a single recv(), goto
+    // back to here for each line
+    //
+processline:
+    eol = strstr(buffer, "\r\n");
+    if (eol == NULL) {
+      y_log_message(Y_LOG_LEVEL_DEBUG, "%d: Haven't found EOL yet", manager->sockfd);
+      continue;
+    }
+
+    // Null terminate each line to be processed individually
+    eol[0] = '\0';
+
+    if (!inmessage) { // Handle system verbs
+      manager->mail_data = mstrcatf(manager->mail_data, "%s\n", buffer);
+      // Replace all lower case letters so verbs are all caps
+      for (i=0; i<4; i++) {
+        if (islower(buffer[i])) {
+          buffer[i] += 'A' - 'a';
+        }
+      }
+      // Null-terminate the verb for strcmp
+      buffer[4] = '\0';
+
+      // Respond to each verb accordingly.
+      // You should replace these with more meaningful
+      // actions than simply printing everything.
+      //
+      if (STREQU(buffer, "HELO")) { // Initial greeting
+        sprintf(bufferout, "250 Ok\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      } else if (STREQU(buffer, "MAIL")) { // New mail from...
+        sprintf(bufferout, "250 Ok\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      } else if (STREQU(buffer, "RCPT")) { // Mail addressed to...
+        sprintf(bufferout, "250 Ok recipient\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      } else if (STREQU(buffer, "DATA")) { // Message contents...
+        sprintf(bufferout, "354 Continue\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+        inmessage = 1;
+      } else if (STREQU(buffer, "RSET")) { // Reset the connection
+        sprintf(bufferout, "250 Ok reset\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      } else if (STREQU(buffer, "NOOP")) { // Do nothing.
+        sprintf(bufferout, "250 Ok noop\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      } else if (STREQU(buffer, "QUIT")) { // Close the connection
+        sprintf(bufferout, "221 Ok\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+        break;
+      } else { // The verb used hasn't been implemented.
+        sprintf(bufferout, "502 Command Not Implemented\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+      }
+    } else { // We are inside the message after a DATA verb.
+      manager->mail_data = mstrcatf(manager->mail_data, "%s\n", buffer);
+      if (STREQU(buffer, ".")) { // A single "." signifies the end
+        sprintf(bufferout, "250 Ok\r\n");
+        send(manager->sockfd, bufferout, strlen(bufferout), 0);
+        //manager->mail_data = mstrcatf(manager->mail_data, bufferout);
+        inmessage = 0;
+      }
+    }
+
+    // Shift the rest of the buffer to the front
+    memmove(buffer, eol+2, BUF_SIZE - (eol + 2 - buffer));
+    buffer_offset -= (eol - buffer) + 2;
+
+    // Do we already have additional lines to process? If so,
+    // commit a horrid sin and goto the line processing section again.
+    if (strstr(buffer, "\r\n")) 
+      goto processline;
+  }
+
+  // All done. Clean up everything and exit.
+  close(manager->sockfd);
+}
+
+static void * simple_smtp(void * args) {
+  struct smtp_manager * manager = (struct smtp_manager *)args;
+  int server_fd; 
+  struct sockaddr_in address; 
+  int opt = 1; 
+  int addrlen = sizeof(address); 
+  
+  if ((server_fd = socket(AF_INET, SOCK_STREAM, 0)) != 0) {
+    if (!setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt))) {
+      address.sin_family = AF_INET;
+      address.sin_addr.s_addr = INADDR_ANY;
+      address.sin_port = htons( manager->port );
+         
+      if (!bind(server_fd, (struct sockaddr *)&address, sizeof(address))) {
+        if (listen(server_fd, 3) >= 0) {
+          if ((manager->sockfd = accept(server_fd, (struct sockaddr *)&address, (socklen_t*)&addrlen)) >= 0) {
+            handle_smtp(manager);
+          } else {
+            y_log_message(Y_LOG_LEVEL_ERROR, "simple_smtp - Error accept");
+          }
+        } else {
+          y_log_message(Y_LOG_LEVEL_ERROR, "simple_smtp - Error listen");
+        }
+      } else {
+        y_log_message(Y_LOG_LEVEL_ERROR, "simple_smtp - Error bind");
+      }
+    } else {
+      y_log_message(Y_LOG_LEVEL_ERROR, "simple_smtp - Error setsockopt");
+    }
+  } else {
+    y_log_message(Y_LOG_LEVEL_ERROR, "simple_smtp - Error socket");
+  }
+
+  pthread_exit(NULL);
+}
+
+static char * get_file_content(const char * file_path) {
+  char * buffer = NULL;
+  size_t length, res;
+  FILE * f;
+
+  f = fopen (file_path, "rb");
+  if (f) {
+    fseek (f, 0, SEEK_END);
+    length = ftell (f);
+    fseek (f, 0, SEEK_SET);
+    buffer = o_malloc((length+1)*sizeof(char));
+    if (buffer) {
+      res = fread (buffer, 1, length, f);
+      if (res != length) {
+        fprintf(stderr, "fread warning, reading %zu while expecting %zu", res, length);
+      }
+      // Add null character at the end of buffer, just in case
+      buffer[length] = '\0';
+    }
+    fclose (f);
+  } else {
+    fprintf(stderr, "error opening file %s\n", file_path);
+  }
+  
+  return buffer;
+}
 
 /**
  * decode a u_map into a string
@@ -226,7 +384,6 @@ int callback_function_multiple_complete(const struct _u_request * request, struc
   return U_CALLBACK_COMPLETE;
 }
 
-
 ssize_t stream_data (void * cls, uint64_t pos, char * buf, size_t max) {
   usleep(100);
   if (pos <= 100) {
@@ -242,8 +399,8 @@ void free_stream_data(void * cls) {
 }
 
 int callback_function_stream (const struct _u_request * request, struct _u_response * response, void * user_data) {
-  ulfius_set_stream_response(response, 200, stream_data, free_stream_data, U_STREAM_SIZE_UNKOWN, 32 * 1024, o_strdup("stream test"));
-  return U_OK;
+  ulfius_set_stream_response(response, 200, stream_data, free_stream_data, U_STREAM_SIZE_UNKNOWN, 32 * 1024, o_strdup("stream test"));
+  return U_CALLBACK_CONTINUE;
 }
 
 size_t my_write_body(void * contents, size_t size, size_t nmemb, void * user_data) {
@@ -261,7 +418,7 @@ int callback_check_utf8_ignored(const struct _u_request * request, struct _u_res
   ck_assert_int_eq(u_map_has_key(request->map_url, "utf8_param_valid1"), 1);
   ck_assert_int_eq(u_map_has_key(request->map_url, "utf8_param_valid2"), 1);
   ck_assert_int_eq(u_map_has_key(request->map_post_body, "utf8_param_valid"), 1);
-  return U_OK;
+  return U_CALLBACK_CONTINUE;
 }
 
 int callback_check_utf8_not_ignored(const struct _u_request * request, struct _u_response * response, void * user_data) {
@@ -273,10 +430,43 @@ int callback_check_utf8_not_ignored(const struct _u_request * request, struct _u
   ck_assert_int_eq(u_map_has_key(request->map_url, "utf8_param_valid1"), 1);
   ck_assert_int_eq(u_map_has_key(request->map_url, "utf8_param_valid2"), 1);
   ck_assert_int_eq(u_map_has_key(request->map_post_body, "utf8_param_valid"), 1);
-  return U_OK;
+  return U_CALLBACK_CONTINUE;
 }
 
-#ifndef U_DISABLE_WEBSOCKET
+int callback_function_position_0(const struct _u_request * request, struct _u_response * response, void * user_data) {
+  ck_assert_int_eq(request->callback_position, 0);
+  return U_CALLBACK_CONTINUE;
+}
+
+int callback_function_position_1(const struct _u_request * request, struct _u_response * response, void * user_data) {
+  ck_assert_int_eq(request->callback_position, 1);
+  return U_CALLBACK_CONTINUE;
+}
+
+int callback_function_position_2(const struct _u_request * request, struct _u_response * response, void * user_data) {
+  ck_assert_int_eq(request->callback_position, 2);
+  return U_CALLBACK_CONTINUE;
+}
+
+int callback_function_redirect_first(const struct _u_request * request, struct _u_response * response, void * user_data) {
+  u_map_put(response->map_header, "Location", "/last");
+  ulfius_set_string_body_response(response, 302, BODY_NOT_REDIRECTED);
+  return U_CALLBACK_CONTINUE;
+}
+
+int callback_function_redirect_last(const struct _u_request * request, struct _u_response * response, void * user_data) {
+  ulfius_set_string_body_response(response, 200, BODY_REDIRECTED);
+  return U_CALLBACK_CONTINUE;
+}
+
+int via_free_with_test = 0;
+
+void free_with_test(void * ptr) {
+  via_free_with_test++;
+  free(ptr);
+}
+
+#ifndef U_DISABLE_GNUTLS
 int callback_auth_client_cert (const struct _u_request * request, struct _u_response * response, void * user_data) {
   char * dn;
   size_t lbuf = 0;
@@ -286,7 +476,13 @@ int callback_auth_client_cert (const struct _u_request * request, struct _u_resp
   dn = o_malloc(lbuf + 1);
   gnutls_x509_crt_get_dn(request->client_cert, dn, &lbuf);
   dn[lbuf] = '\0';
-  ck_assert_str_eq(dn, "C=CA,ST=Quebec,L=Quebec,O=Ulfius,OU=test-client,CN=localhost,EMAIL=webmaster@localhost");
+  ck_assert_ptr_ne(o_strstr(dn, "C=CA"), NULL);
+  ck_assert_ptr_ne(o_strstr(dn, "ST=Quebec"), NULL);
+  ck_assert_ptr_ne(o_strstr(dn, "L=Quebec"), NULL);
+  ck_assert_ptr_ne(o_strstr(dn, "O=Ulfius"), NULL);
+  ck_assert_ptr_ne(o_strstr(dn, "OU=test-client"), NULL);
+  ck_assert_ptr_ne(o_strstr(dn, "CN=localhost"), NULL);
+  ck_assert_ptr_ne(o_strstr(dn, "EMAIL=webmaster@localhost"), NULL);
   ck_assert_int_eq(ulfius_set_string_body_response(response, 200, dn), U_OK);
   o_free(dn);
 
@@ -440,6 +636,144 @@ START_TEST(test_ulfius_simple_endpoint)
 }
 END_TEST
 
+#if MHD_VERSION >= 0x00095208
+START_TEST(test_ulfius_net_type_endpoint)
+{
+  struct _u_instance u_instance;
+  struct _u_request request;
+  struct _u_response response;
+  struct sockaddr_in ipv4addr;
+  struct sockaddr_in6 ipv6addr;
+  
+  // Test network accepting IPV4 and IPV6 connections
+  ck_assert_int_eq(ulfius_init_instance_ipv6(&u_instance, 8080, NULL, U_USE_ALL, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "empty", NULL, 0, &callback_function_empty, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://localhost:8080/empty");
+  request.network_type = U_USE_IPV4;
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://[::1]:8080/empty");
+  request.network_type = U_USE_IPV6;
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+  
+  // Test network accepting IPV6 only connections
+  ck_assert_int_eq(ulfius_init_instance_ipv6(&u_instance, 8080, NULL, U_USE_IPV6, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "empty", NULL, 0, &callback_function_empty, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://127.0.0.1:8080/empty");
+  request.network_type = U_USE_IPV4;
+  ck_assert_int_ne(ulfius_send_http_request(&request, NULL), U_OK);
+  ulfius_clean_request(&request);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://[::1]:8080/empty");
+  request.network_type = U_USE_IPV6;
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+  
+  // Test network accepting IPV4 only connections
+  ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, NULL, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "empty", NULL, 0, &callback_function_empty, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://localhost:8080/empty");
+  request.network_type = U_USE_IPV4;
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://[::1]:8080/empty");
+  request.network_type = U_USE_IPV6;
+  ck_assert_int_ne(ulfius_send_http_request(&request, NULL), U_OK);
+  ulfius_clean_request(&request);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+  
+  // Test network binding to an IPV4 address
+  memset(&ipv4addr, 0, sizeof(ipv4addr));
+  ipv4addr.sin_family = AF_INET;
+  ipv4addr.sin_port = htons(8080);
+  ipv4addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+  ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, &ipv4addr, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "empty", NULL, 0, &callback_function_empty, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://localhost:8080/empty");
+  request.network_type = U_USE_IPV4;
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://[::1]:8080/empty");
+  request.network_type = U_USE_IPV6;
+  ck_assert_int_ne(ulfius_send_http_request(&request, NULL), U_OK);
+  ulfius_clean_request(&request);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+  
+  // Test network binding to an IPV6 address
+  memset(&ipv6addr, 0, sizeof(ipv6addr));
+  ipv6addr.sin6_family = AF_INET6;
+  ipv6addr.sin6_port = htons(8080);
+  ipv6addr.sin6_addr = in6addr_loopback;
+  ck_assert_int_eq(ulfius_init_instance_ipv6(&u_instance, 8080, &ipv6addr, U_USE_IPV6, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "empty", NULL, 0, &callback_function_empty, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://127.0.0.1:8080/empty");
+  request.network_type = U_USE_IPV4;
+  ck_assert_int_ne(ulfius_send_http_request(&request, NULL), U_OK);
+  ulfius_clean_request(&request);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://[::1]:8080/empty");
+  request.network_type = U_USE_IPV6;
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+}
+END_TEST
+#endif
+
 START_TEST(test_ulfius_endpoint_parameters)
 {
   struct _u_instance u_instance;
@@ -730,17 +1064,174 @@ START_TEST(test_ulfius_utf8_ignored)
 }
 END_TEST
 
-#ifndef U_DISABLE_WEBSOCKET
+START_TEST(test_ulfius_endpoint_callback_position)
+{
+  struct _u_instance u_instance;
+  struct _u_request request;
+  struct _u_response response;
+  
+  ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, NULL, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "callback_position", "*", 0, &callback_function_position_0, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "callback_position", "*", 0, &callback_function_position_1, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "callback_position", "*", 0, &callback_function_position_2, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://localhost:8080/callback_position");
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+}
+END_TEST
+
+START_TEST(test_ulfius_MHD_set_response_with_other_free)
+{
+  struct _u_instance u_instance;
+  struct _u_request request;
+  struct _u_response response;
+  
+  o_set_alloc_funcs(&malloc, &realloc, &free_with_test);
+  ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, NULL, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "/", NULL, 0, &callback_function_return_url, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://localhost:8080/");
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+  
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+  // We can only test that via_free_with_test has been incremented.
+  // The only way to know for sure now is to verify with valgrind and the memory consumption
+  ck_assert_int_gt(via_free_with_test, 1);
+}
+END_TEST
+
+#define PORT_PLAIN 2525
+#define PORT_RICH 2526
+#define FROM "from"
+#define TO "to"
+#define CC "cc"
+#define BCC "bcc"
+#define CONTENT_TYPE "text/ulfius; charset=utf-42"
+#define SUBJECT "subject"
+#define BODY "mail body"
+
+START_TEST(test_ulfius_send_smtp)
+{
+  pthread_t thread;
+  struct smtp_manager manager;
+
+  manager.mail_data = NULL;
+  manager.port = PORT_PLAIN;
+  manager.sockfd = 0;
+
+  ck_assert_int_eq(ulfius_send_smtp_email("localhost", PORT_PLAIN, 0, 0, NULL, NULL, FROM, TO, CC, BCC, SUBJECT, BODY), U_ERROR_LIBCURL);
+  ck_assert_int_eq(ulfius_send_smtp_email(NULL, PORT_PLAIN, 0, 0, NULL, NULL, FROM, TO, CC, BCC, SUBJECT, BODY), U_ERROR_PARAMS);
+  ck_assert_int_eq(ulfius_send_smtp_email("localhost", PORT_PLAIN, 0, 0, NULL, NULL, NULL, TO, CC, BCC, SUBJECT, BODY), U_ERROR_PARAMS);
+  ck_assert_int_eq(ulfius_send_smtp_email("localhost", PORT_PLAIN, 0, 0, NULL, NULL, FROM, NULL, CC, BCC, SUBJECT, BODY), U_ERROR_PARAMS);
+  ck_assert_int_eq(ulfius_send_smtp_email("localhost", PORT_PLAIN, 0, 0, NULL, NULL, FROM, TO, CC, BCC, SUBJECT, NULL), U_ERROR_PARAMS);
+  
+  pthread_create(&thread, NULL, simple_smtp, &manager);
+  ck_assert_int_eq(ulfius_send_smtp_email("localhost", PORT_PLAIN, 0, 0, NULL, NULL, FROM, TO, CC, BCC, SUBJECT, BODY), U_OK);
+  pthread_join(thread, NULL);
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "HELO"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "MAIL FROM:<" FROM ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "RCPT TO:<" TO ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "RCPT TO:<" CC ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "RCPT TO:<" BCC ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "Subject: " SUBJECT));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, BODY));
+  o_free(manager.mail_data);
+  manager.mail_data = NULL;
+}
+END_TEST
+
+START_TEST(test_ulfius_send_rich_smtp)
+{
+  pthread_t thread;
+  struct smtp_manager manager;
+
+  manager.mail_data = NULL;
+  manager.port = PORT_RICH;
+  manager.sockfd = 0;
+
+  pthread_create(&thread, NULL, simple_smtp, &manager);
+  ck_assert_int_eq(ulfius_send_smtp_rich_email("localhost", PORT_RICH, 0, 0, NULL, NULL, FROM, TO, CC, BCC, CONTENT_TYPE, SUBJECT, BODY), U_OK);
+  pthread_join(thread, NULL);
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "HELO"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "MAIL FROM:<" FROM ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "RCPT TO:<" TO ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "RCPT TO:<" CC ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "RCPT TO:<" BCC ">"));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "Content-Type: " CONTENT_TYPE));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, "Subject: " SUBJECT));
+  ck_assert_ptr_ne(NULL, o_strstr(manager.mail_data, BODY));
+  o_free(manager.mail_data);
+  manager.mail_data = NULL;
+}
+END_TEST
+
+START_TEST(test_ulfius_follow_redirect)
+{
+  struct _u_instance u_instance;
+  struct _u_request request;
+  struct _u_response response;
+  
+  ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, NULL, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "first", "*", 0, &callback_function_redirect_first, NULL), U_OK);
+  ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "last", "*", 0, &callback_function_redirect_last, NULL), U_OK);
+  ck_assert_int_eq(ulfius_start_framework(&u_instance), U_OK);
+  
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("http://localhost:8080/first");
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 302);
+  ck_assert_str_eq(u_map_get(response.map_header, "Location"), "/last");
+  ck_assert_int_eq(o_strncmp(response.binary_body, BODY_NOT_REDIRECTED, o_strlen(BODY_NOT_REDIRECTED)), 0);
+  ulfius_clean_response(&response);
+  
+  ulfius_init_response(&response);
+  request.follow_redirect = 1;
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(o_strncmp(response.binary_body, BODY_REDIRECTED, o_strlen(BODY_REDIRECTED)), 0);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_response(&response);
+  
+  ulfius_clean_request(&request);
+  ulfius_stop_framework(&u_instance);
+  ulfius_clean_instance(&u_instance);
+}
+END_TEST
+
+#ifndef U_DISABLE_GNUTLS
 START_TEST(test_ulfius_server_ca_trust)
 {
   struct _u_instance u_instance;
-  
+  char * cert = get_file_content("cert/server.crt"), * key = get_file_content("cert/server.key"), * ca = get_file_content("cert/root1.crt");
+
+  ck_assert_ptr_ne(cert, NULL);
+  ck_assert_ptr_ne(key, NULL);
+  ck_assert_ptr_ne(ca, NULL);
   ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, NULL, NULL), U_OK);
   ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "stream", NULL, 0, &callback_function_stream, NULL), U_OK);
-  ck_assert_int_ne(ulfius_start_secure_ca_trust_framework(&u_instance, NULL, CERT_PEM, CERT_CA), U_OK);
-  ck_assert_int_ne(ulfius_start_secure_ca_trust_framework(&u_instance, CERT_KEY, NULL, CERT_CA), U_OK);
-  ck_assert_int_eq(ulfius_start_secure_ca_trust_framework(&u_instance, CERT_KEY, CERT_PEM, CERT_CA), U_OK);
+  ck_assert_int_ne(ulfius_start_secure_ca_trust_framework(&u_instance, NULL, cert, ca), U_OK);
+  ck_assert_int_ne(ulfius_start_secure_ca_trust_framework(&u_instance, key, NULL, ca), U_OK);
+  ck_assert_int_eq(ulfius_start_secure_ca_trust_framework(&u_instance, key, cert, ca), U_OK);
   
+  o_free(cert);
+  o_free(key);
+  o_free(ca);
   ulfius_stop_framework(&u_instance);
   ulfius_clean_instance(&u_instance);
 }
@@ -751,16 +1242,20 @@ START_TEST(test_ulfius_client_certificate)
   struct _u_instance u_instance;
   struct _u_request request;
   struct _u_response response;
+  char * cert = get_file_content("cert/server.crt"), * key = get_file_content("cert/server.key"), * ca = get_file_content("cert/root1.crt");
   
+  ck_assert_ptr_ne(cert, NULL);
+  ck_assert_ptr_ne(key, NULL);
+  ck_assert_ptr_ne(ca, NULL);
   ck_assert_int_eq(ulfius_init_instance(&u_instance, 8080, NULL, NULL), U_OK);
   ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "cert_auth", NULL, 0, &callback_auth_client_cert, NULL), U_OK);
   ck_assert_int_eq(ulfius_add_endpoint_by_val(&u_instance, "GET", "nocert_auth", NULL, 0, &callback_no_auth_client_cert, NULL), U_OK);
-  ck_assert_int_eq(ulfius_start_secure_ca_trust_framework(&u_instance, CERT_KEY, CERT_PEM, CERT_CA), U_OK);
+  ck_assert_int_eq(ulfius_start_secure_ca_trust_framework(&u_instance, key, cert, ca), U_OK);
 
   ulfius_init_request(&request);
   request.http_url = o_strdup("https://localhost:8080/nocert_auth");
   request.check_server_certificate = 0;
-  request.client_key_file = o_strdup("client.key");
+  request.client_key_file = o_strdup("cert/client1.key");
   ulfius_init_response(&response);
   ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
   ck_assert_int_eq(response.status, 200);
@@ -770,24 +1265,51 @@ START_TEST(test_ulfius_client_certificate)
   ulfius_init_request(&request);
   request.http_url = o_strdup("https://localhost:8080/nocert_auth");
   request.check_server_certificate = 0;
-  request.client_cert_file = o_strdup("client.crt");
+  request.client_cert_file = o_strdup("cert/client1.crt");
   ulfius_init_response(&response);
   ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
   ck_assert_int_eq(response.status, 200);
   ulfius_clean_request(&request);
   ulfius_clean_response(&response);
   
+  // Test with a valid certificate signed by the expected CA (root1)
   ulfius_init_request(&request);
   request.http_url = o_strdup("https://localhost:8080/cert_auth");
   request.check_server_certificate = 0;
-  request.client_cert_file = o_strdup("client.crt");
-  request.client_key_file = o_strdup("client.key");
+  request.client_cert_file = o_strdup("cert/client1.crt");
+  request.client_key_file = o_strdup("cert/client1.key");
   ulfius_init_response(&response);
   ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
   ck_assert_int_eq(response.status, 200);
   ulfius_clean_request(&request);
   ulfius_clean_response(&response);
   
+  // Test with a valid certificate signed by another CA (root2)
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("https://localhost:8080/nocert_auth");
+  request.check_server_certificate = 0;
+  request.client_cert_file = o_strdup("cert/client2.crt");
+  request.client_key_file = o_strdup("cert/client2.key");
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ck_assert_int_eq(response.status, 200);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+
+  // Test with a certificate self-signed (client3)
+  ulfius_init_request(&request);
+  request.http_url = o_strdup("https://localhost:8080/nocert_auth");
+  request.check_server_certificate = 0;
+  request.client_cert_file = o_strdup("cert/client3.crt");
+  request.client_key_file = o_strdup("cert/client3.key");
+  ulfius_init_response(&response);
+  ck_assert_int_eq(ulfius_send_http_request(&request, &response), U_OK);
+  ulfius_clean_request(&request);
+  ulfius_clean_response(&response);
+
+  o_free(cert);
+  o_free(key);
+  o_free(ca);
   ulfius_stop_framework(&u_instance);
   ulfius_clean_instance(&u_instance);
 }
@@ -802,13 +1324,21 @@ static Suite *ulfius_suite(void)
   s = suite_create("Ulfius framework function tests");
   tc_core = tcase_create("test_ulfius_framework");
   tcase_add_test(tc_core, test_ulfius_simple_endpoint);
+#if MHD_VERSION >= 0x00095208
+  tcase_add_test(tc_core, test_ulfius_net_type_endpoint);
+#endif
   tcase_add_test(tc_core, test_ulfius_endpoint_parameters);
   tcase_add_test(tc_core, test_ulfius_endpoint_injection);
   tcase_add_test(tc_core, test_ulfius_endpoint_multiple);
   tcase_add_test(tc_core, test_ulfius_endpoint_stream);
   tcase_add_test(tc_core, test_ulfius_utf8_not_ignored);
   tcase_add_test(tc_core, test_ulfius_utf8_ignored);
-#ifndef U_DISABLE_WEBSOCKET
+  tcase_add_test(tc_core, test_ulfius_endpoint_callback_position);
+  tcase_add_test(tc_core, test_ulfius_MHD_set_response_with_other_free);
+  tcase_add_test(tc_core, test_ulfius_send_smtp);
+  tcase_add_test(tc_core, test_ulfius_send_rich_smtp);
+  tcase_add_test(tc_core, test_ulfius_follow_redirect);
+#ifndef U_DISABLE_GNUTLS
   tcase_add_test(tc_core, test_ulfius_server_ca_trust);
   tcase_add_test(tc_core, test_ulfius_client_certificate);
 #endif
@@ -823,7 +1353,8 @@ int main(int argc, char *argv[])
   int number_failed;
   Suite *s;
   SRunner *sr;
-  //y_init_logs("Ulfius", Y_LOG_MODE_CONSOLE, Y_LOG_LEVEL_DEBUG, NULL, "Starting Ulfius core tests");
+  //y_init_logs("Ulfius", Y_LOG_MODE_CONSOLE, Y_LOG_LEVEL_DEBUG, NULL, "Starting Ulfius framework tests");
+  ulfius_global_init();
   s = ulfius_suite();
   sr = srunner_create(s);
 
@@ -831,6 +1362,7 @@ int main(int argc, char *argv[])
   number_failed = srunner_ntests_failed(sr);
   srunner_free(sr);
   
+  ulfius_global_close();
   //y_close_logs();
   return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
 }
